/* Custom asciidoc styles */

body {
  margin: 1em 2em;
  font-family: verdana, sans-serif;
  font-size: 12px;
}
div.sectionbody, #author, #revnumber, #revdate, #revremark, #toctitle, #footer {
  font-family: verdana, sans-serif;
}
a, a:visited, a:hover, a:active {
  color: #00c;
  text-decoration: none;
}
em {
  color: #252;
}
strong {
  color: #000;
}
tt {
  color: #366;
}
#author {
  color: #555;
}
#toc {
  line-height: 1.4em;
  padding-bottom: 3em;
}
#toctitle {
  margin-bottom: .6em;
  color: #555;
}
#toc .toclevel0, #toc .toclevel1 {
  display: block;
  font-size: 1em;
}
#toc .toclevel2, #toc .toclevel3, #toc .toclevel4 {
  display: block;
  font-size: 1em;
}
h1, h2, h3, h4, h5, h6 {
  margin: 1.2em 0;
}
h1 {
  padding: 1.2em;
  border: 1px solid #c0d0ff;
  border-radius: 8px;
  box-shadow: 1px 1px 3px #aac;
  background: #ddeaee;
  background: -moz-linear-gradient(#bbddff, #eefaff);
  background: -webkit-linear-gradient(#bbddff, #eefaff);
  background: linear-gradient(#bbddff, #eefaff);
  color: #224b8d;
  font-size: 2.2em;
  text-align: center;
}
h2 {
  padding: .5em;
  border: 1px solid #d0d0d0;
  border-radius: 6px;
  box-shadow: 1px 1px 2px #bbb;
  background: #f2f2f2;
  background: -moz-linear-gradient(#e0e0e0, #f6f6f6);
  background: -webkit-linear-gradient(#e0e0e0, #f6f6f6);
  background: linear-gradient(#e0e0e0, #f6f6f6);
  color: #444;
  font-size: 1.7em;
}
h3 {
  border: none;
  margin: .7em 0 1em 0;
  color: #555;
  font-size: 1.5em;
}
h4 {
  color: #555;
  font-size: 1.3em;
}
h5 {
  color: #555;
  font-size: 1.2em;
}
span.command {
  padding: .2em .5em;
  border: 1px solid #ccc;
  border-radius: 4px;
  box-shadow: 1px 1px 2px #ccc;
  margin-right: 12px;
  background: #eee;
  font-size: 1.1em;
}
p.table {
  margin-top: 0;
  margin-bottom: 0;
}
div.listingblock > div.content {
  border: 1px solid #e2e2e2;
  border-radius: 4px;
  background: #fbfbfb;
}
.listingblock pre {
  white-space: pre-wrap;
  word-wrap: break-word;
}
table th {
  background: #fbfbfb;
  color: #000;
  font-size: 1.1em;
}
table.tableblock, th.tableblock, td.tableblock, .tableblock table th, .tableblock table td {
  border: 1px solid #bbb;
}
table.tableblock {
  border: 2px solid #888;
}
div.title {
  color: #525252;
}
div.admonitionblock .icon {
  text-decoration: none;
}
dt.hdlist1 {
  margin-bottom: .5em;
  color: #252;
}
.ulist {
  word-wrap: break-word;
}
.key {
  padding: .1em .4em;
  border-top: 1px solid #d5d5e5;
  border-left: 1px solid #d5d5e5;
  border-bottom: 1px solid #a5a5b5;
  border-right: 1px solid #a5a5b5;
  border-radius: 3px;
  box-shadow: 1px 1px 3px #d5d5e5;
  margin: 0 2px;
  background: #eef;
  font-family: monospace;
  font-size: 1.1em;
  line-height: 1.6em;
}
.key.ctrl {
  border-top: 1px solid #dfd0d0;
  border-left: 1px solid #dfd0d0;
  border-bottom: 1px solid #afa0a0;
  border-right: 1px solid #afa0a0;
  border-radius: 3px;
  box-shadow: 1px 1px 3px #dfd0d0;
  background: #fff0f0;
  text-transform: capitalize;
}
.key.alt {
  border-top: 1px solid #dfc5b4;
  border-left: 1px solid #dfc5b4;
  border-bottom: 1px solid #af9484;
  border-right: 1px solid #af9484;
  border-radius: 3px;
  box-shadow: 1px 1px 3px #dfc5b4;
  background: #ffe4d4;
  text-transform: capitalize;
}
.key.shift {
  border-top: 1px solid #d5d5d5;
  border-left: 1px solid #d5d5d5;
  border-bottom: 1px solid #a5a5a5;
  border-right: 1px solid #a5a5a5;
  border-radius: 3px;
  box-shadow: 1px 1px 3px #d5d5d5;
  background: #f5f5f5;
  text-transform: capitalize;
}
.key.other {
  border-top: 1px solid #d0e0d0;
  border-left: 1px solid #d0e0d0;
  border-bottom: 1px solid #a0b0a0;
  border-right: 1px solid #a0b0a0;
  border-radius: 3px;
  box-shadow: 1px 1px 3px #d0e0d0;
  background: #eafaea;
}
.keyplus {
  font-size: .8em;
}
.hexa {
  padding: 0 .2em;
  border: 1px solid #88b;
  margin: 0 .2em;
  color: #007;
  font-family: monospace;
  font-size: .85em;
}
#footer {
  border-top: solid 1px silver;
}

/* Asciidoc styles with table of contents on the left (toc2) */

body {
  margin-left: 26em;
  margin-right: 2em;
}
#toc {
  width: 23em;
}
#toc .toclevel2 {
  margin-left: 1.5em;
  font-size: .9em;
}
#toc .toclevel3 {
  margin-left: 3em;
  font-size: .9em;
}
#toc .toclevel4 {
  margin-left: 4.5em;
  font-size: .9em;
}
