/*
 * test-url.cpp - test URL functions
 *
 * Copyright (C) 2014-2016 Sébastien Helleu <flashcode@flashtux.org>
 *
 * This file is part of WeeChat, the extensible chat client.
 *
 * WeeChat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * WeeChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WeeChat.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "CppUTest/TestHarness.h"

extern "C"
{
#include "src/core/wee-url.h"
}

TEST_GROUP(Url)
{
};

/*
 * Tests functions:
 *   weeurl_download
 */

TEST(Url, Download)
{
    /* TODO: write tests */
}

/*
 * Tests functions:
 *   weeurl_option_add_to_infolist
 */

TEST(Url, AddToInfolist)
{
    /* TODO: write tests */
}
